from .model.ctabgan import CTAB_XTRA_DP


from .model import evaluation as _evaluation

from .datasets.dataset_loader import load_demo , display_demo


__all__ = [
    "CTAB_XTRA_DP",
    "evaluation",
    "load_demo",
    "display_demo",
]


__version__ = "0.1.0"


import numpy as np
import pandas as pd
from sklearn import preprocessing
from sklearn import model_selection

from .Log_preparation import Log_preparation
from .Categorical_preparation import Categorical_preparation

class DataPrep(object):
  
    def __init__(self, raw_df: pd.DataFrame, categorical_list, log_list):
        
    
        self.preprossesers =  self.setup_preprocesses(raw_df,categorical_list, log_list)
        super().__init__()

    
    def setup_preprocesses(self, data, categorical_list, log_list):
        
        preprocesses = []
        for column in data.columns:
            column_type = self.get_preprocesses_transform(data,column, categorical_list, log_list)
            preprocesses.append(column_type)
        return preprocesses


    def get_preprocesses_transform(self, data,column, categorical_list, log_list):
        if column in categorical_list: return Categorical_preparation(data[column])
        if column in log_list: return Log_preparation(data[column])
        return None  # No transformation specified

    
    def preprocesses_transform(self,data):
        assert len(self.preprossesers) == data.shape[1]
        data = data.copy()
        for idx, preprosseser in enumerate(self.preprossesers):
            if preprosseser is None: continue
            column = data.iloc[:,idx]
            data.iloc[:,idx] = preprosseser.transform(column)
            
        return data

    def preprocesses_inverse_transform(self,data):
        assert len(self.preprossesers) == data.shape[1]
        data = data.copy()
        for idx, preprosseser in enumerate(self.preprossesers):
            if preprosseser is None: continue
            column = data.iloc[:,idx]
            data.iloc[:,idx] = preprosseser.inverse_transform(column)
            
        return data

    def get_label_encoded(self,column_index, conditioning_value):
        # Takes in column index and find the encoded values of the conditioning value
        preprosess = self.preprossesers[column_index]
        if not isinstance(preprosess, Categorical_preparation): raise ValueError("Column is not categorical")
        return preprosess.get_label_encoded(conditioning_value)
        



    
                
  
        




import numpy as np

from scipy.sparse import issparse






def gower_distance(data_x, data_y=None, weight=None, mix_features=None):
    X = data_x
    Y = data_y if data_y is not None else data_x

    

    if not isinstance(X, np.ndarray): X = np.asarray(X)
    if not isinstance(Y, np.ndarray): Y = np.asarray(Y)

    if X.shape[1] != Y.shape[1]:
        raise ValueError(f"X and Y must have the same number of columns")
    n_cols = X.shape[1] # the number of columns for both X and Y
    
    # Set default weights if not provided
    if weight is None:
        weight = np.ones(n_cols)
    else:
        weight = np.asarray(weight)

    XY = np.concatenate((X, Y))
    XY = replace_nulls_with_placeholder(XY)
 
    if issparse(X) or issparse(Y): raise TypeError("Sparse matrices are not supported!") 

    numeric_array, numeric_mask = build_numeric_array_and_mask(XY)
    column_max, column_min, column_means, column_stds = compute_column_statistics(numeric_array, numeric_mask)
    column_range = column_max - column_min

    XY = np.where(
        numeric_mask,
        (numeric_array - column_min) / column_range,
        XY  #
    )

    n_rows_X = X.shape[0]
    X_proc = XY[:n_rows_X]
    Y_proc = XY[n_rows_X:] 
    n_rows_X = X_proc.shape[0]
    n_rows_Y = Y_proc.shape[0]
    
    numeric_mask = numeric_mask.astype(bool)
    numeric_mask_X = numeric_mask[:n_rows_X]
    numeric_mask_Y = numeric_mask[n_rows_X:]

    total_weighted = np.zeros((n_rows_X, n_rows_Y))

    for i in range(n_rows_X):
        xi = X_proc[i]
        i_numeric_mask = numeric_mask_X[i]
        

        common_ones = np.logical_and(i_numeric_mask, numeric_mask_Y)
        common_zeros = np.logical_and(~i_numeric_mask, ~numeric_mask_Y)
        differing_values = np.logical_xor(i_numeric_mask, numeric_mask_Y)

        
        
        numerical_abs = np.abs(np.where(common_ones,Y_proc,np.nan) - np.where(common_ones,xi,np.nan))
        weighted_numeric_abs = numerical_abs * np.where(common_ones, weight, 0)
        weighted_numeric_abs_sum = np.nansum(weighted_numeric_abs, axis=1)
        
        categorical = np.where(
            common_zeros,
            xi != Y_proc,
            False
        )
        weighted_categorical = categorical * np.where(common_zeros, weight, 0)
        weighted_categorical_sum = np.nansum(weighted_categorical, axis=1)

        differing_values = differing_values * np.where(differing_values, weight, 0)
        weighted_differing_values_sum = np.nansum(differing_values, axis=1)

        total_weighted_sum = weighted_numeric_abs_sum + weighted_categorical_sum + weighted_differing_values_sum
        total_weighted_row = total_weighted_sum / np.sum(weight)

        total_weighted[i] = total_weighted_row

        
    return total_weighted
        




def build_numeric_array_and_mask(array):
    array = np.asarray(array, dtype=object)  # Handle mixed types safely
    n_rows, n_cols = array.shape

    numeric_array = np.zeros((n_rows, n_cols), dtype=float)
    mask = np.zeros((n_rows, n_cols), dtype=int)  # 0 = categorical by default


    for i in range(n_rows):
        for j in range(n_cols):
            val = array[i, j]

            if isinstance(val, (int, float)) and not isinstance(val, bool):
                numeric_array[i, j] = val
                mask[i, j] = 1
            elif isinstance(val, str):
                try:
                    float_value = float(val)
                    numeric_array[i, j] = float_value
                    mask[i, j] = 1
                except:
                    pass
            # else: keep it 0 (categorical)
    
    return numeric_array, mask



def compute_column_statistics(numeric_array, numeric_mask):
 
    column_max = []
    column_min = []
    column_means = []
    column_stds = []
    for col_idx in range(numeric_array.shape[1]):
        # Get the mask for this column
        col_mask = numeric_mask[:, col_idx]
        
        # Get the values for this column where mask is 1
        masked_values = numeric_array[:, col_idx][col_mask == 1]
        
        max_val = np.max(masked_values) if len(masked_values) > 0 else np.nan
        min_val = np.min(masked_values) if len(masked_values) > 0 else np.nan
        mean_val = np.mean(masked_values) if len(masked_values) > 0 else np.nan
        std_val = np.std(masked_values) if len(masked_values) > 0 else np.nan
        
        column_max.append(max_val)
        column_min.append(min_val)
        column_means.append(mean_val)
        column_stds.append(std_val)

    column_max = np.array(column_max)
    column_min = np.array(column_min)
    column_means = np.array(column_means)
    column_stds = np.array(column_stds)
    return column_max, column_min, column_means, column_stds




def is_null(val):
    return val is None or (isinstance(val, float) and np.isnan(val))

def replace_nulls_with_placeholder(matrix, placeholder="__MISSING__"):
    matrix = np.asarray(matrix, dtype=object)
    return np.array([
        [placeholder if is_null(x) else x for x in row]
        for row in matrix
    ], dtype=object)





import numpy as np
from sklearn import preprocessing

class Categorical_preparation:
    def __init__(self, column):
        self.label_encoder = preprocessing.LabelEncoder()
        self.label_encoder.fit(column)
        
    def transform(self, column):
      
        encoded_column = self.label_encoder.transform(column)
        return np.asarray(encoded_column,int)

    def inverse_transform(self, column):
        column = np.asarray(column,int)
        return self.label_encoder.inverse_transform(column)

    def get_label_encoded(self, conditioning_value):
        # returns the encoded value of the real condition value
        return self.label_encoder.transform([conditioning_value])[0]


import numpy as np
import pandas as pd

class Log_preparation:
    def __init__(self, column,eps=1):
        column_numpy = column.to_numpy()
        lower_value = np.nanmin(column_numpy)
        self.lower_bound = 0 if lower_value > 0 else (eps if lower_value == 0 else -lower_value + eps) # Set lower bound to ensure all values are positive
        

    def transform(self, column):
        column_numpy = column.to_numpy()
        return np.log(column_numpy + self.lower_bound)

    def inverse_transform(self, column):
        column_numpy = column.to_numpy()
        return np.exp(column_numpy) - self.lower_bound




import numpy as np

class Null_value_transformer:
    def __init__(self):
        pass

    def fit(self, data, categorical_columns, mixed_columns):
        data = data.copy()
        self.nan_placeholder = "__NAN_PLACEHOLDER__"
        self.nan_columns = []
        for column in data.columns:
            data_col = data[column]
            if not data_col.isnull().any(): continue
            self.nan_columns.append(column)
            data[column] = data_col.fillna(self.nan_placeholder)
            if column not in categorical_columns:
                # Append the nan placeholder to the mixed columns such the method knows how to handle it
                if column in mixed_columns:
                    mixed_columns[column].append(self.nan_placeholder)
                else:
                    mixed_columns[column] = [self.nan_placeholder]
        return mixed_columns

        
        
    def transform(self,data):
        data = data.copy()
        for column in self.nan_columns:
            data[column] = data[column].replace(np.nan, self.nan_placeholder)
        return data

    def inverse_transform(self,data):
        data = data.copy()
        for column in self.nan_columns:
            data[column] = data[column].replace(self.nan_placeholder,np.nan)
        return data



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""RDP analysis of the Sampled Gaussian Mechanism.

Functionality for computing Renyi differential privacy (RDP) of an additive
Sampled Gaussian Mechanism (SGM). Its public interface consists of two methods:
  compute_rdp(q, noise_multiplier, T, orders) computes RDP for SGM iterated
                                   T times.
  get_privacy_spent(orders, rdp, target_eps, target_delta) computes delta
                                   (or eps) given RDP at multiple orders and
                                   a target value for eps (or delta).

Example use:

Suppose that we have run an SGM applied to a function with l2-sensitivity 1.
Its parameters are given as a list of tuples (q1, sigma1, T1), ...,
(qk, sigma_k, Tk), and we wish to compute eps for a given delta.
The example code would be:

  max_order = 32
  orders = range(2, max_order + 1)
  rdp = np.zeros_like(orders, dtype=float)
  for q, sigma, T in parameters:
   rdp += rdp_accountant.compute_rdp(q, sigma, T, orders)
  eps, _, opt_order = rdp_accountant.get_privacy_spent(rdp, target_delta=delta)
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import sys

import numpy as np
from scipy import special
import scipy
import six

########################
# LOG-SPACE ARITHMETIC #
########################


def _log_add(logx, logy):
  """Add two numbers in the log space."""
  a, b = min(logx, logy), max(logx, logy)
  if a == -np.inf:  # adding 0
    return b
  # Use exp(a) + exp(b) = (exp(a - b) + 1) * exp(b)
  return math.log1p(math.exp(a - b)) + b  # log1p(x) = log(x + 1)


def _log_sub(logx, logy):
  """Subtract two numbers in the log space. Answer must be non-negative."""
  if logx < logy:
    raise ValueError("The result of subtraction must be non-negative.")
  if logy == -np.inf:  # subtracting 0
    return logx
  if logx == logy:
    return -np.inf  # 0 is represented as -np.inf in the log space.

  try:
    # Use exp(x) - exp(y) = (exp(x - y) - 1) * exp(y).
    return math.log(math.expm1(logx - logy)) + logy  # expm1(x) = exp(x) - 1
  except OverflowError:
    return logx


def _log_print(logx):
  """Pretty print."""
  if logx < math.log(sys.float_info.max):
    return "{}".format(math.exp(logx))
  else:
    return "exp({})".format(logx)


def _compute_log_a_int(q, sigma, alpha):
  """Compute log(A_alpha) for integer alpha. 0 < q < 1."""
  assert isinstance(alpha, six.integer_types)

  # Initialize with 0 in the log space.
  log_a = -np.inf

  for i in range(alpha + 1):
    log_coef_i = (
        math.log(special.binom(alpha, i)) + i * math.log(q) +
        (alpha - i) * math.log(1 - q))

    s = log_coef_i + (i * i - i) / (2 * (sigma**2))
    log_a = _log_add(log_a, s)

  return float(log_a)


def _compute_log_a_int_optimized(q, sigma, alpha):
    """Optimized computation of log(A_alpha) for integer alpha. 0 < q < 1."""
    assert isinstance(alpha, six.integer_types)
    if not (0 < q < 1):
        raise ValueError("q has to be in the open interval (0, 1).")
    
    # Create array of all i values at once
    i_values = np.arange(alpha + 1)
    
    # Compute binomial coefficients using logarithms of factorials
    log_binom = (scipy.special.gammaln(alpha + 1) - 
                 scipy.special.gammaln(i_values + 1) - 
                 scipy.special.gammaln(alpha - i_values + 1))
    
    # Compute all terms in the sum at once
    log_q = np.log(q)
    log_1_q = np.log(1 - q)
    log_coef = log_binom + i_values * log_q + (alpha - i_values) * log_1_q
    
    # Compute the quadratic term
    quadratic_term = (i_values * i_values - i_values) / (2 * (sigma**2))
    
    # Combine terms
    terms = log_coef + quadratic_term
    
    # Use scipy's logsumexp which is numerically stable
    log_a = scipy.special.logsumexp(terms)
    
    return float(log_a)


def _compute_log_a_frac(q, sigma, alpha):
  """Compute log(A_alpha) for fractional alpha. 0 < q < 1."""
  # The two parts of A_alpha, integrals over (-inf,z0] and [z0, +inf), are
  # initialized to 0 in the log space:
  log_a0, log_a1 = -np.inf, -np.inf
  i = 0

  z0 = sigma**2 * math.log(1 / q - 1) + .5

  while True:  # do ... until loop
    coef = special.binom(alpha, i)
    log_coef = math.log(abs(coef))
    j = alpha - i

    log_t0 = log_coef + i * math.log(q) + j * math.log(1 - q)
    log_t1 = log_coef + j * math.log(q) + i * math.log(1 - q)

    log_e0 = math.log(.5) + _log_erfc((i - z0) / (math.sqrt(2) * sigma))
    log_e1 = math.log(.5) + _log_erfc((z0 - j) / (math.sqrt(2) * sigma))

    log_s0 = log_t0 + (i * i - i) / (2 * (sigma**2)) + log_e0
    log_s1 = log_t1 + (j * j - j) / (2 * (sigma**2)) + log_e1

    if coef > 0:
      log_a0 = _log_add(log_a0, log_s0)
      log_a1 = _log_add(log_a1, log_s1)
    else:
      log_a0 = _log_sub(log_a0, log_s0)
      log_a1 = _log_sub(log_a1, log_s1)

    i += 1
    if max(log_s0, log_s1) < -30:
      break

  return _log_add(log_a0, log_a1)


def _compute_log_a(q, sigma, alpha):
  """Compute log(A_alpha) for any positive finite alpha."""
  if float(alpha).is_integer():
    return _compute_log_a_int_optimized(q, sigma, int(alpha))
  else:
    return _compute_log_a_frac(q, sigma, alpha)


def _log_erfc(x):
  """Compute log(erfc(x)) with high accuracy for large x."""
  try:
    return math.log(2) + special.log_ndtr(-x * 2**.5)
  except NameError:
    # If log_ndtr is not available, approximate as follows:
    r = special.erfc(x)
    if r == 0.0:
      # Using the Laurent series at infinity for the tail of the erfc function:
      #     erfc(x) ~ exp(-x^2-.5/x^2+.625/x^4)/(x*pi^.5)
      # To verify in Mathematica:
      #     Series[Log[Erfc[x]] + Log[x] + Log[Pi]/2 + x^2, {x, Infinity, 6}]
      return (-math.log(math.pi) / 2 - math.log(x) - x**2 - .5 * x**-2 +
              .625 * x**-4 - 37. / 24. * x**-6 + 353. / 64. * x**-8)
    else:
      return math.log(r)


def _compute_delta(orders, rdp, eps):
  """Compute delta given a list of RDP values and target epsilon.

  Args:
    orders: An array (or a scalar) of orders.
    rdp: A list (or a scalar) of RDP guarantees.
    eps: The target epsilon.

  Returns:
    Pair of (delta, optimal_order).

  Raises:
    ValueError: If input is malformed.

  """
  orders_vec = np.atleast_1d(orders)
  rdp_vec = np.atleast_1d(rdp)

  if len(orders_vec) != len(rdp_vec):
    raise ValueError("Input lists must have the same length.")

  deltas = np.exp((rdp_vec - eps) * (orders_vec - 1))
  idx_opt = np.argmin(deltas)
  return min(deltas[idx_opt], 1.), orders_vec[idx_opt]


def _compute_eps(orders, rdp, delta):
  """Compute epsilon given a list of RDP values and target delta.

  Args:
    orders: An array (or a scalar) of orders.
    rdp: A list (or a scalar) of RDP guarantees.
    delta: The target delta.

  Returns:
    Pair of (eps, optimal_order).

  Raises:
    ValueError: If input is malformed.

  """
  orders_vec = np.atleast_1d(orders)
  rdp_vec = np.atleast_1d(rdp)

  if len(orders_vec) != len(rdp_vec):
    raise ValueError("Input lists must have the same length.")

  eps = rdp_vec - math.log(delta) / (orders_vec - 1)

  idx_opt = np.nanargmin(eps)  # Ignore NaNs
  return eps[idx_opt], orders_vec[idx_opt]


def _compute_rdp(q, sigma, alpha):
  """Compute RDP of the Sampled Gaussian mechanism at order alpha.

  Args:
    q: The sampling rate.
    sigma: The std of the additive Gaussian noise.
    alpha: The order at which RDP is computed.

  Returns:
    RDP at alpha, can be np.inf.
  """
  if q == 0:
    return 0

  if q == 1.:
    return alpha / (2 * sigma**2)

  if np.isinf(alpha):
    return np.inf

  return _compute_log_a(q, sigma, alpha) / (alpha - 1)


def compute_rdp(q, noise_multiplier, steps, orders): # TODO: Remove all dependence to this and instead use the single
                                                     # The single does not need to be recompeuted for each step
  """Compute RDP of the Sampled Gaussian Mechanism.

  Args:
    q: The sampling rate.
    noise_multiplier: The ratio of the standard deviation of the Gaussian noise
        to the l2-sensitivity of the function to which it is added.
    steps: The number of steps.
    orders: An array (or a scalar) of RDP orders.

  Returns:
    The RDPs at all orders, can be np.inf.
  """
  if np.isscalar(orders):
    rdp = _compute_rdp(q, noise_multiplier, orders)
  else:
    rdp = np.array([_compute_rdp(q, noise_multiplier, order)
                    for order in orders])

  return rdp * steps

def compute_rdp_single(q, noise_multiplier, orders):
  """Compute RDP of the Sampled Gaussian Mechanism.

  Args:
    q: The sampling rate.
    noise_multiplier: The ratio of the standard deviation of the Gaussian noise
        to the l2-sensitivity of the function to which it is added.
    orders: An array (or a scalar) of RDP orders.

  Returns:
    The RDPs at all orders, can be np.inf.
  """
  if np.isscalar(orders):
    rdp = _compute_rdp(q, noise_multiplier, orders)
  else:
    rdp = np.array([_compute_rdp(q, noise_multiplier, order)
                    for order in orders])

  return rdp


def get_privacy_spent(orders, rdp, target_eps=None, target_delta=None):
  """Compute delta (or eps) for given eps (or delta) from RDP values.

  Args:
    orders: An array (or a scalar) of RDP orders.
    rdp: An array of RDP values. Must be of the same length as the orders list.
    target_eps: If not None, the epsilon for which we compute the corresponding
              delta.
    target_delta: If not None, the delta for which we compute the corresponding
              epsilon. Exactly one of target_eps and target_delta must be None.

  Returns:
    eps, delta, opt_order.

  Raises:
    ValueError: If target_eps and target_delta are messed up.
  """
  if target_eps is None and target_delta is None:
    raise ValueError(
        "Exactly one out of eps and delta must be None. (Both are).")

  if target_eps is not None and target_delta is not None:
    raise ValueError(
        "Exactly one out of eps and delta must be None. (None is).")

  if target_eps is not None:
    delta, opt_order = _compute_delta(orders, rdp, target_eps)
    return target_eps, delta, opt_order
  else:
    eps, opt_order = _compute_eps(orders, rdp, target_delta)
    return eps, target_delta, opt_order


def compute_rdp_from_ledger(ledger, orders):
  """Compute RDP of Sampled Gaussian Mechanism from ledger.

  Args:
    ledger: A formatted privacy ledger.
    orders: An array (or a scalar) of RDP orders.

  Returns:
    RDP at all orders, can be np.inf.
  """
  total_rdp = np.zeros_like(orders, dtype=float)
  for sample in ledger:
    # Compute equivalent z from l2_clip_bounds and noise stddevs in sample.
    # See https://arxiv.org/pdf/1812.06210.pdf for derivation of this formula.
    effective_z = sum([
        (q.noise_stddev / q.l2_norm_bound)**-2 for q in sample.queries])**-0.5
    total_rdp += compute_rdp(
        sample.selection_probability, effective_z, 1, orders)
  return total_rdp



def find_sigma_for_target_epsilon(q, steps, target_epsilon, target_delta, 
                                  precision=0.01, max_iterations=100,verbose=False):
    """
    Find the noise multiplier (sigma) required to achieve a target epsilon with given delta.
    
    Args:
        q: The sampling rate (batch_size / dataset_size)
        steps: The total number of training steps
        target_epsilon: The target privacy budget
        target_delta: The target delta parameter (typically 1/N or smaller)
        precision: Desired precision for sigma (default: 0.01)
        max_iterations: Maximum number of binary search iterations (default: 100)
        
    Returns:
        The sigma value that achieves the target privacy guarantee
    """
    # Initial search range for sigma
    sigma_min, sigma_max = 0.01, 100.0
    
    # Initialize tracking variables
    iteration = 0
    best_sigma = None
    best_eps_diff = float('inf')
    
    # Set RDP orders
    orders = range(2, 4096)
    
    # Binary search for the appropriate sigma
    while sigma_max - sigma_min > precision and iteration < max_iterations:
        # Try middle point of current range
        sigma_mid = (sigma_min + sigma_max) / 2
        
        # Calculate RDP for these parameters
        rdp = compute_rdp_single(q, sigma_mid, orders) * steps
        # Convert RDP to (ε, δ)-DP
        eps, _, opt_order = get_privacy_spent(orders, rdp, target_delta=target_delta)
        
        # Track closest approximation
        eps_diff = abs(eps - target_epsilon)
        if eps_diff < best_eps_diff:
            best_sigma = sigma_mid
            best_eps_diff = eps_diff
        
        
        # Adjust search range
        if eps > target_epsilon:  # Current sigma is too small (privacy too weak)
            sigma_min = sigma_mid
        else:  # Current sigma is too large (privacy too strong)
            sigma_max = sigma_mid
        
        iteration += 1
    
    # If binary search didn't converge, use best approximation
    if iteration == max_iterations:
        
        return best_sigma
    
    # Calculate final epsilon to verify
    final_rdp = compute_rdp(q, sigma_max, steps, orders)
    final_eps, _, _ = get_privacy_spent(orders, final_rdp, target_delta=target_delta)
    
    if verbose: print(f"Found sigma={sigma_max:.4f} for epsilon={final_eps:.4f} (target was {target_epsilon})")
    return sigma_max  # Return slightly conservative estimate




import numpy as np
import pandas as pd 
from sklearn import metrics
from sklearn import model_selection
from sklearn.preprocessing import MinMaxScaler,StandardScaler, LabelEncoder

from sklearn.neural_network import MLPClassifier
from sklearn.linear_model import LogisticRegression, LinearRegression, Ridge, Lasso, BayesianRidge
from sklearn import svm,tree
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor ,HistGradientBoostingClassifier, HistGradientBoostingRegressor
from dython.nominal import associations
from scipy.stats import wasserstein_distance
from scipy.spatial import distance
import warnings
from .gower_mix import gower_distance
from collections import defaultdict

from xgboost import XGBClassifier, XGBRegressor
from lightgbm import LGBMClassifier, LGBMRegressor
from catboost import CatBoostClassifier, CatBoostRegressor


warnings.filterwarnings("ignore")





def get_summary_metrics(real,
                        fake,
                        categorical=[],
                        mixed={},
                        mnar=True,
                        problem="classification"):

    simmlarity = stat_sim(real,fake,categorical,mixed,mnar)
    if problem == "regression":
      models_to_run = ["dt","rf","hgb","cat","xgb"]
    elif problem == "classification":
      models_to_run = ["lr","dt","rf","xgb"]
    else:
      raise ValueError("Problem type must be either 'classification' or 'regression'")
    utility = get_utility_metrics(real,[fake],categorical=categorical, mixed=mixed,scaler="MinMax",problem=problem, models=models_to_run,test_ratio=.20)
    privacy = privacy_metrics(real,fake,metric='gower')
    return utility, simmlarity, privacy


def supervised_model_training(x_train, y_train, x_test, 
                              y_test, model_name,problem_type):
  
  model = get_supervised_model(model_name,problem_type)
  
  model.fit(x_train, y_train)
  pred = model.predict(x_test)

  if problem_type == "classification":
    acc = metrics.accuracy_score(y_test,pred)*100
    if len(np.unique(y_train))>2: # We have mulitclass classification
      predict = model.predict_proba(x_test)        
      auc = metrics.roc_auc_score(y_test, predict,average="weighted",multi_class="ovr")
      f1_score = metrics.precision_recall_fscore_support(y_test, pred,average="weighted")[2]

    else: # We have binary classification
      predict = model.predict_proba(x_test)[:,1]    
      auc = metrics.roc_auc_score(y_test, predict)
      f1_score = metrics.precision_recall_fscore_support(y_test,pred)[2].mean()
    return [acc, auc, f1_score] 
  
  if problem_type == "regression":
    mse = metrics.mean_absolute_percentage_error(y_test,pred)
    evs = metrics.explained_variance_score(y_test, pred)
    r2_score = metrics.r2_score(y_test,pred)
    return [mse, evs, r2_score]

  raise ValueError(f"Unknown problem type: {problem_type}. Supported types are 'classification' and 'regression'.")




def get_supervised_model(model_name,problem_type,random_state=42):
  if problem_type == "classification":
    if model_name == 'lr': return LogisticRegression(random_state=random_state,max_iter=500)
    if model_name == 'svm': return svm.SVC(random_state=random_state,probability=True)

    if model_name == 'dt': return tree.DecisionTreeClassifier(random_state=random_state)
    if model_name == 'rf': return RandomForestClassifier(random_state=random_state)
    if model_name == 'mlp': return MLPClassifier(random_state=random_state,max_iter=100)
    if model_name == 'xgb': return XGBClassifier(random_state=random_state, use_label_encoder=False, eval_metric='logloss')
    if model_name == 'lgbm': return LGBMClassifier(random_state=random_state,verbose=-1)
    if model_name == 'cat': return CatBoostClassifier(random_state=random_state, verbose=0)
    raise ValueError(f"Unknown model name: {model_name} for classification")
  
  if problem_type == "regression":
    if model_name == 'lin': return LinearRegression()
    if model_name == 'ridge': return Ridge(random_state=random_state)
    if model_name == 'lasso': return Lasso(random_state=random_state)
    if model_name == 'B_ridge': return BayesianRidge()
  
    if model_name == 'dt': return tree.DecisionTreeRegressor(random_state=random_state)
    if model_name == 'rf': return RandomForestRegressor(random_state=random_state)
    if model_name == 'hgb': return HistGradientBoostingRegressor(random_state=random_state)
    if model_name == 'xgb': return XGBRegressor(random_state=random_state)
    if model_name == 'lgbm': return LGBMRegressor(random_state=random_state,verbose=-1)
    if model_name == 'cat': return CatBoostRegressor(random_state=random_state, verbose=0)
    raise ValueError(f"Unknown model name: {model_name} for regression")

  raise ValueError(f"Unknown problem type: {problem_type}. Supported types are 'classification' and 'regression'.")




def split_data(df, test_ratio=.20, target=None,problem="classification"):
  if problem == "classification":
    if target is None: raise ValueError("Target column must be specified for classification")
    train, test = model_selection.train_test_split(df, test_size=test_ratio, stratify=df[target], random_state=42)
  else:
    train, test = model_selection.train_test_split(df, test_size=test_ratio, random_state=42)
  return train, test


def get_utility_metrics(real,
                        fakes,
                        categorical=[],
                        mixed={},
                        target=None,
                        scaler="MinMax",
                        problem="classification", 
                        models=["lr","dt","rf","mlp"],
                        test_ratio=.20):


    real = real.copy()
    categorical = categorical.copy()

    if target is None:
      target = real.columns[-1]
    
    real, additional_categorical_cols, _=_process_mixed_columns(real, mixed, continuous_placeholder=np.nan)

    categorical.extend(additional_categorical_cols)

    # Encode categorical columns in the real dataset
    for col in categorical:
      label_encoder = LabelEncoder()
      real[col] = label_encoder.fit_transform(real[col])
      
    data_real_y = real.loc[:, target].to_numpy()
    data_real_X = real.drop(target,axis=1).to_numpy()

    data_dim = real.shape[1]
  
    if problem == "classification":
      X_train_real, X_test_real, y_train_real, y_test_real = model_selection.train_test_split(data_real_X ,data_real_y, test_size=test_ratio, stratify=data_real_y,random_state=42) 
    else:
      X_train_real, X_test_real, y_train_real, y_test_real = model_selection.train_test_split(data_real_X ,data_real_y, test_size=test_ratio,random_state=42) 
    
    # Apply scaling
    if scaler=="MinMax":
        scaler_real = MinMaxScaler()
    else:
        scaler_real = StandardScaler()
    

    
    scaler_real.fit(X_train_real)
    X_train_real_scaled = scaler_real.transform(X_train_real)
    X_test_real_scaled = scaler_real.transform(X_test_real)



    if problem == "classification":
      metrics_list = ["Accuracy", "AUC", "F1-Score"]
    else:
      metrics_list = ["MAPE", "Explained Variance", "R2 Score"]


    all_real_results = []
    for model in models:
      real_results = supervised_model_training(X_train_real_scaled,y_train_real,X_test_real_scaled,y_test_real,model,problem)
      all_real_results.append(real_results)
      
    all_fake_results_avg = []
    
    for fake in fakes:
      fake = fake.copy()

      fake, _, _=_process_mixed_columns(fake, mixed, continuous_placeholder=np.nan)

      # Encode categorical columns in the fake dataset
      for col in categorical:
        label_encoder = LabelEncoder()
        fake[col] = label_encoder.fit_transform(fake[col])

      data_fake = fake.to_numpy()

      data_fake_y = fake.loc[:, target].to_numpy()
      data_fake_X = fake.drop(target,axis=1).to_numpy()
     
      
      
      if problem=="classification":
        X_train_fake, _ , y_train_fake, _ = model_selection.train_test_split(data_fake_X ,data_fake_y, test_size=test_ratio, stratify=data_fake_y,random_state=42) 
      else:
        X_train_fake, _ , y_train_fake, _ = model_selection.train_test_split(data_fake_X ,data_fake_y, test_size=test_ratio,random_state=42)  


      # Apply scaling
      if scaler=="MinMax":
        scaler_fake = MinMaxScaler()
      else:
        scaler_fake = StandardScaler()
      
      scaler_fake.fit(data_fake_X)
      
      X_train_fake_scaled = scaler_fake.transform(X_train_fake)
      
      all_fake_results = []
      for model in models:
        fake_results = supervised_model_training(X_train_fake_scaled,y_train_fake,X_test_real_scaled,y_test_real,model,problem)
        all_fake_results.append(fake_results)

      all_fake_results_avg.append(all_fake_results)
    

    

    
    real_results_df = pd.DataFrame(all_real_results, columns=metrics_list, index=models)

    fake_results_dfs = []

    for fake_results in all_fake_results_avg:
      fake_df = pd.DataFrame(fake_results, columns=metrics_list, index=models)
      fake_results_dfs.append(fake_df)

  # Return real and list of fake DataFrames
    return real_results_df, fake_results_dfs




def stat_sim(real,fake,categorical=[],mixed={},mnar=True):
    

    nan_placeholder = "__MISSING__"
    continuous_placeholder='__CONTINUOUS__'


    real = real.copy()
    fake = fake.copy()
    categorical = categorical.copy()

   
    mixed = defaultdict(list, mixed)

    columns_with_nan = []
    if mnar: # If Missing not at random, we replace NaN values with a placeholder,effectively treating them as a separate category
      
      nan_counts_real = real.isna().sum()
      nan_counts_fake = fake.isna().sum()
      columns_with_nan_real = nan_counts_real[nan_counts_real > 0].index.tolist()
      columns_with_nan_fake = nan_counts_fake[nan_counts_fake > 0].index.tolist()

      columns_with_nan = list(set(columns_with_nan_real + columns_with_nan_fake))

      for column in columns_with_nan:
        if continuous_placeholder not in mixed[column]: mixed[column].append(nan_placeholder)

      real = real.fillna(nan_placeholder)
      fake = fake.fillna(nan_placeholder)

    real_processed, additional_categorical_cols, additional_continuous_cols = _process_mixed_columns(real, mixed, continuous_placeholder=continuous_placeholder)
    fake_processed, _, _ = _process_mixed_columns(fake, mixed, continuous_placeholder=continuous_placeholder)
    categorical.extend(additional_categorical_cols)


    corr_diff = alternative_correlation(real_processed, fake_processed, columns_to_remove=additional_continuous_cols) 
    summary, column_stats = column_similarity(real_processed, fake_processed, categorical)

    return summary, column_stats, corr_diff



def column_similarity(real, fake, categorical=[]):


    real = real.copy()
    fake = fake.copy()

    for col in categorical:
      label_encoder = LabelEncoder()
      real[col] = label_encoder.fit_transform(real[col])
      fake[col] = label_encoder.transform(fake[col])


    column_stats = []
    default_columns_weight = 1
    weights = []
    for column in real.columns:
        
        if column in categorical:
            
            real_pdf=(real[column].value_counts()/real[column].value_counts().sum())
            fake_pdf=(fake[column].value_counts()/fake[column].value_counts().sum())
            categories = (fake[column].value_counts()/fake[column].value_counts().sum()).keys().tolist()
            sorted_categories = sorted(categories)
            
            real_pdf_values = [] 
            fake_pdf_values = []

            for i in sorted_categories:
                #if i not in real_pdd: raise ValueError(f"Category {i} present in fake but not in real data")
                if i not in real_pdf: real_pdf[i] = 0 #TODO: might not be like this
                if i not in fake_pdf: fake_pdf[i] = 0
                real_pdf_values.append(real_pdf[i])
                fake_pdf_values.append(fake_pdf[i])
            
                
            js_distance = (distance.jensenshannon(real_pdf_values,fake_pdf_values, 2.0))
 
            statistics = [column, "JSD",js_distance, default_columns_weight]
    
            weights.append(default_columns_weight)
    
        else:
            scaler = MinMaxScaler()
            scaler.fit(real[column].values.reshape(-1,1))
            l1 = scaler.transform(real[column].values.reshape(-1,1)).flatten()
            l2 = scaler.transform(fake[column].values.reshape(-1,1)).flatten()
            weight = default_columns_weight
            #TODO: this might drop stuff it should not
            weight = 1 - np.isnan(l1).sum()/len(l1)
            l1 = l1[~np.isnan(l1)]
            l2 = l2[~np.isnan(l2)]
            w_distance = (wasserstein_distance(l1,l2))
            statistics = [column, "WD", w_distance, weight]
            weights.append(weight)
        column_stats.append(statistics)

    column_stats = pd.DataFrame(column_stats, columns=["Column", "Metric", "Distance", "Weight"])

    summary = column_stats.groupby('Metric').agg({
        'Distance': [
            ('Weighted_Avg', lambda x: np.average(x, weights=column_stats.loc[x.index, 'Weight'])),
            ('Mean', 'mean'),
            ('Std', 'std')
        ]
    })
    summary.columns = summary.columns.get_level_values(1)
    summary = summary.reset_index()
    return summary, column_stats

def alternative_correlation(real, fake, columns_to_remove=[],difference_metric="mae"):
    """
    Calculate the correlation between real and fake datasets
    
    Args:
        real (pd.DataFrame): Real dataset
        fake (pd.DataFrame): Fake dataset
        columns_to_remove (list): List of columns to remove from the datasets
    
    Returns:
        float: Correlation distance
    """
    # Remove specified columns
    real = real.drop(columns=columns_to_remove, errors='ignore')
    fake = fake.drop(columns=columns_to_remove, errors='ignore')

    # Calculate correlation matrices
    real_corr = associations(real, compute_only=True)["corr"]
    fake_corr = associations(fake, compute_only=True)["corr"]

    if difference_metric == "mae":
        # Calculate absolute difference between correlation matrices
        diff_matrix = np.abs(real_corr - fake_corr)
        columnwise_avg_diff = diff_matrix.mean(axis=0)
        return columnwise_avg_diff.mean()
    
    if difference_metric == "forbenious":
        # Calculate Frobenius norm of the difference between correlation matrices
        return np.linalg.norm(real_corr - fake_corr, ord='fro')
    
    raise ValueError(f"Unknown difference metric: {difference_metric}. Supported metrics are 'mae' and 'forbenious'.")
    
    

def _process_mixed_columns(df, mixed, continuous_placeholder='__CONTINUOUS__'):
    """
    Process mixed columns by extracting categorical values
    
    Args:
        df (pd.DataFrame): Input DataFrame
        mixed (dict): Dictionary of mixed columns with their categories
        continuous_placeholder (str): Placeholder to indicate continuous value in categorical part of mixed columns
    
    Returns:
        tuple: (processed_df, additional_categorical_cols)
    """
    # Create a copy of the DataFrame
    df_copy = df.copy()
    additional_categorical_cols = []
    additional_continuous_cols = []
    
    for col, categories in mixed.items():
        # Create a new categorical column
        new_cat_col_name = f"{col}_categorical"
        new_cont_col_name = f"{col}_continuous"
        
        # Create a series for categorical values
        categorical_series = df_copy[col].apply(
            lambda x: x if x in categories else continuous_placeholder
        )
        
        # Add the new categorical column
        df_copy[new_cat_col_name] = categorical_series
        additional_categorical_cols.append(new_cat_col_name)

        continuous_series = df_copy[col].apply(
            lambda x: x if x not in categories else np.nan
        )
        df_copy[new_cont_col_name] = continuous_series
        additional_continuous_cols.append(new_cont_col_name)

        # Drop the original column
        df_copy.drop(columns=[col], inplace=True)
    
    return df_copy, additional_categorical_cols, additional_continuous_cols





def privacy_metrics(real, 
                    fake, 
                    metric = 'gower', 
                    data_percent=15,
                    verbose=False):
    """
    Calculate privacy metrics between real and fake datasets.
    
    Parameters:
    real (DataFrame): Real dataset
    fake (DataFrame): Synthetic dataset
    data_percent (int): Percentage of data to sample for analysis
    
    Returns:
    dict: Dictionary containing detailed privacy metrics
    """
    import numpy as np
    from sklearn.preprocessing import StandardScaler
    from sklearn import metrics
    import pandas as pd
    
    # Create a metrics dictionary to store all results
    privacy_summary = []
    
    # Sample the data
    real_refined = real.sample(n=int(len(real)*(.01*data_percent)), random_state=42).to_numpy()
    fake_refined = fake.sample(n=int(len(fake)*(.01*data_percent)), random_state=42).to_numpy()
    
    if verbose:
      print("==== Privacy Metrics Analysis ====")
      print(f"Using {data_percent}% of data for analysis")
      
      
      print(f"Real data sample size: {len(real_refined)}")
      print(f"Fake data sample size: {len(fake_refined)}")
      

      print("\nCalculating pairwise distances...")
    
    # Compute distance matrix from real-fake, real-real, and fake-fake
    dist_rf, dist_rr, dist_ff = _calculate_pariwise_distances(real_refined, fake_refined, metric=metric)

    # Remove the diagonal to ignore self-distances
    dist_rr = dist_rr[~np.eye(dist_rr.shape[0],dtype=bool)].reshape(dist_rr.shape[0],-1)
    dist_ff = dist_ff[~np.eye(dist_ff.shape[0],dtype=bool)].reshape(dist_ff.shape[0],-1) 

    
    
    # Find smallest two distances for each metric
    smallest_two_indexes_rf = [dist_rf[i].argsort()[:2] for i in range(len(dist_rf))]
    smallest_two_rf = [dist_rf[i][smallest_two_indexes_rf[i]] for i in range(len(dist_rf))]       
    smallest_two_indexes_rr = [dist_rr[i].argsort()[:2] for i in range(len(dist_rr))]
    smallest_two_rr = [dist_rr[i][smallest_two_indexes_rr[i]] for i in range(len(dist_rr))]
    smallest_two_indexes_ff = [dist_ff[i].argsort()[:2] for i in range(len(dist_ff))]
    smallest_two_ff = [dist_ff[i][smallest_two_indexes_ff[i]] for i in range(len(dist_ff))]
    
    # Cacluaate nearest neighbor distance ratios
    nndr_ratio_rr = np.array([i[0]/i[1] for i in smallest_two_rr])
    nndr_ratio_ff = np.array([i[0]/i[1] for i in smallest_two_ff])
    nndr_ratio_rf = np.array([i[0]/i[1] for i in smallest_two_rf])

    # Calculate distance to closest record
    dcr_rf = np.array([i[0] for i in smallest_two_rf])
    dcr_rr = np.array([i[0] for i in smallest_two_rr])
    dcr_ff = np.array([i[0] for i in smallest_two_ff])
    
    # We report the lowerst 5th percentile of the metrics
    nndr_fifth_perc_rr = np.percentile(nndr_ratio_rr, 5)
    nndr_fifth_perc_ff = np.percentile(nndr_ratio_ff, 5)
    nndr_fifth_perc_rf = np.percentile(nndr_ratio_rf, 5)
    
    fifth_perc_rf = np.percentile(dcr_rf, 5)
    fifth_perc_rr = np.percentile(dcr_rr, 5)
    fifth_perc_ff = np.percentile(dcr_ff, 5)
    
    

    privacy_percentiles = pd.DataFrame({
    'rr': [nndr_fifth_perc_rr, fifth_perc_rr],
    'ff': [nndr_fifth_perc_ff, fifth_perc_ff],
    'rf': [nndr_fifth_perc_rf, fifth_perc_rf]
    }, index=['nndr_5th', 'dcr_5th'])

    privacy_summary.append(privacy_percentiles)
    
    if verbose:
      print("\n== Nearest Neighbor Ratio Metrics (5th percentile) ==")
      print(f"Real-to-Real NN Ratio (5th): {nndr_fifth_perc_rr:.4f}")
      print(f"Fake-to-Fake NN Ratio (5th): {nndr_fifth_perc_ff:.4f}")
      print(f"Real-to-Fake NN Ratio (5th): {nndr_fifth_perc_rf:.4f}")
    
      print("\n== Minimum Distance Metrics (5th percentile) ==")
      print(f"Real-to-Real Min Distance (5th): {fifth_perc_rr:.4f}")
      print(f"Fake-to-Fake Min Distance (5th): {fifth_perc_ff:.4f}")
      print(f"Real-to-Fake Min Distance (5th): {fifth_perc_rf:.4f}")


    nnaa = compute_nnaa(dist_rf,dist_rr,dist_ff)
    privacy_summary.append(nnaa)

    if verbose:
      print("\n== Nearest Neighbor Adversarial Accuracy (NNAA) ==")
      print(f"NNAA: {nnaa:.4f}")
    
    # Calculate privacy risk score
    # Higher score indicates lower privacy risk
    privacy_risk = (fifth_perc_rf / (fifth_perc_rr + fifth_perc_ff) * 2)
    privacy_summary.append(privacy_risk)
    
    if verbose:
      print(f"\nPrivacy Risk Score: {privacy_risk:.4f}")
      print("(Higher score indicates better privacy protection)")
    
    
    
    
    if verbose:
      print("\n==== Summary Interpretation ====")
      if fifth_perc_rf > (fifth_perc_rr + fifth_perc_ff)/2:
          print("✓ Good distance between real and synthetic records")
      else:
          print("⚠ Synthetic records may be too similar to real data")
          
      if nndr_fifth_perc_rf > (nndr_fifth_perc_rr + nndr_fifth_perc_ff)/2:
          print("✓ Good neighbor distance ratios")
      else:
          print("⚠ Neighbor distance ratios indicate possible privacy concerns")
      
    
    return privacy_summary


# Compute the Nearest Neighbor Adversarial Accuaracy (NNAA) between real and fake data
def compute_nnaa(dist_rf, # Distance matrix between real and fake data
                 dist_rr, # Distance matrix between real and real data
                 dist_ff): # Distance matrix between fake and fake data

    real_data_size = dist_rf.shape[0] # Number of real data points
    fake_data_size = dist_rf.shape[1] # Number of fake data points

    dist_fr = dist_rf.T # Transpose to get distance matrix between fake and real data

    dist_rf_min = np.min(dist_rf, axis=1) # Minimum distance from real to fake
    dist_fr_min = np.min(dist_fr, axis=1) # Minimum distance from fake to real
    dist_rr_min = np.min(dist_rr, axis=1) # Minimum distance from real to real
    dist_ff_min = np.min(dist_ff, axis=1) # Minimum distance from fake to fake

    real_data_closest = (dist_rr_min < dist_rf_min).astype(int) # The closest point to real data is reals data
    fake_data_closest = (dist_ff_min < dist_fr_min).astype(int) # The closest point to fake data is fakes data

    nnaa = (real_data_closest.mean() + fake_data_closest.mean()) / 2 # Take the mean of both and average them

    return nnaa



def _calculate_pariwise_distances(real, fake, metric='gower'):
    dist_rf = 0 # Pairwise distance between real and fake
    dist_rr = 0 # Pairwise distance in real data
    dist_ff = 0 # Pairwise distance between fake and fake
    if metric == 'gower':
      dist_rf = gower_distance(real,fake)
      dist_rr = gower_distance(real,real)
      dist_ff = gower_distance(fake,fake)
      return dist_rf, dist_rr, dist_ff


    if metric == 'euclidean':
      scalerR = StandardScaler()
      scalerF = StandardScaler()
      scalerR.fit(real)
      scalerF.fit(fake)
      df_real_scaled = scalerR.transform(real)
      df_fake_scaled = scalerF.transform(fake)

      dist_rf = metrics.pairwise_distances(df_real_scaled, Y=df_fake_scaled, metric='minkowski', n_jobs=-1)
      dist_rr = metrics.pairwise_distances(df_real_scaled, Y=None, metric='minkowski', n_jobs=-1)
      dist_ff = metrics.pairwise_distances(df_fake_scaled, Y=None, metric='minkowski', n_jobs=-1)

      return dist_rf, dist_rr, dist_ff
    

    raise ValueError(f"Unknown metric: {metric}. Supported metrics are 'gower' and 'euclidean'.")




import matplotlib.pyplot as plt


def compare_dataframes(
    df: pd.DataFrame,
    syn: pd.DataFrame,
    nominal_values: dict,
    categorical_columns: list = None,
    columns: list = None
):
    # 100 % ChatGPTs work
    """
    Compare each column in df and syn using:
    - Histogram if the column has continuous data
    - Bar chart if the column has categorical data

    Mixed columns are plotted alone; one-type columns (cat or cont only) are plotted two per row.

    Parameters:
    - df: Original DataFrame
    - syn: Synthetic DataFrame (same columns)
    - nominal_values: dict mapping column names to categorical entries (can include np.nan)
    - categorical_columns: list of columns to treat as fully categorical (optional)
    - columns: optional list of column names to include in the plot
    """
    categorical_columns = categorical_columns or []
    columns_to_plot = columns if columns is not None else df.columns
    one_part_buffer = []

    def plot_one_part_pair(pair):
        plt.figure(figsize=(14, 5))
        for i, (col, cat_vals, cont_vals, is_cat) in enumerate(pair):
            plt.subplot(1, 2, i + 1)
            if is_cat:
                all_categories = set(cat_vals[0].unique()).union(set(cat_vals[1].unique()))
                orig_counts = cat_vals[0].value_counts().reindex(all_categories, fill_value=0)
                syn_counts = cat_vals[1].value_counts().reindex(all_categories, fill_value=0)
                width = 0.35
                x = np.arange(len(all_categories))
                plt.bar(x - width/2, orig_counts.values, width=width, label='Original', edgecolor='black')
                plt.bar(x + width/2, syn_counts.values, width=width, label='Synthetic', edgecolor='black')
                plt.xticks(ticks=x, labels=all_categories, rotation=45)
                plt.title(f'{col} (Categorical)')
                plt.xlabel('Category')
                plt.ylabel('Count')
                plt.legend()
            else:
                bins = np.histogram_bin_edges(np.concatenate([cont_vals[0], cont_vals[1]]), bins=50)
                plt.hist(cont_vals[0], bins=bins, alpha=0.6, label='Original', edgecolor='black')
                plt.hist(cont_vals[1], bins=bins, alpha=0.6, label='Synthetic', edgecolor='black')
                plt.title(f'{col} (Continuous)')
                plt.xlabel('Value')
                plt.ylabel('Frequency')
                plt.legend()
        plt.tight_layout()
        plt.show()

    for col in columns_to_plot:
        original = df[col]
        synthetic = syn[col]

        if col in categorical_columns:
            cat_orig = original.fillna('NaN').astype(str)
            cat_syn = synthetic.fillna('NaN').astype(str)
            cont_orig = cont_syn = pd.Series(dtype=float)
        else:
            mixed_values = nominal_values.get(col, [])

            def split_series(series, mixed_values):
                is_categorical = series.apply(
                    lambda x: any((pd.isna(val) and pd.isna(x)) or (str(x) == str(val)) for val in mixed_values)
                )
                categorical = series[is_categorical].fillna('NaN')
                continuous = pd.to_numeric(series[~is_categorical], errors='coerce').dropna().astype(float)
                return categorical, continuous

            cat_orig, cont_orig = split_series(original, mixed_values)
            cat_syn, cont_syn = split_series(synthetic, mixed_values)

        has_cont = not cont_orig.empty or not cont_syn.empty
        has_cat = not cat_orig.empty or not cat_syn.empty

        # Case 1: Mixed → full-row layout
        if has_cat and has_cont:
            plt.figure(figsize=(14, 6))
            plt.subplot(1, 2, 1)
            bins = np.histogram_bin_edges(np.concatenate([cont_orig, cont_syn]), bins=50)
            plt.hist(cont_orig, bins=bins, alpha=0.6, label='Original', edgecolor='black')
            plt.hist(cont_syn, bins=bins, alpha=0.6, label='Synthetic', edgecolor='black')
            plt.title(f'Histogram (Continuous): {col}')
            plt.xlabel('Value')
            plt.ylabel('Frequency')
            plt.legend()

            plt.subplot(1, 2, 2)
            all_categories = set(cat_orig.unique()).union(set(cat_syn.unique()))
            orig_counts = cat_orig.value_counts().reindex(all_categories, fill_value=0)
            syn_counts = cat_syn.value_counts().reindex(all_categories, fill_value=0)
            width = 0.35
            x = np.arange(len(all_categories))
            plt.bar(x - width/2, orig_counts.values, width=width, label='Original', edgecolor='black')
            plt.bar(x + width/2, syn_counts.values, width=width, label='Synthetic', edgecolor='black')
            plt.xticks(ticks=x, labels=all_categories, rotation=45)
            plt.title(f'Bar Chart (Categorical): {col}')
            plt.xlabel('Category')
            plt.ylabel('Count')
            plt.legend()
            plt.tight_layout()
            plt.show()

        # Case 2: Only categorical or continuous → buffer it for side-by-side plotting
        elif has_cat:
            one_part_buffer.append((col, (cat_orig, cat_syn), None, True))
        elif has_cont:
            one_part_buffer.append((col, None, (cont_orig, cont_syn), False))

        # Plot in pairs if buffer has two items
        if len(one_part_buffer) == 2:
            plot_one_part_pair(one_part_buffer)
            one_part_buffer = []

    # Final leftover plot if only one item in buffer
    if len(one_part_buffer) == 1:
        plot_one_part_pair([one_part_buffer[0]])




import warnings

class Data_type_assigner:
    def __init__(self, data,integer_columns = [],mixed_columns = {}):
    
        self.data_type = data.dtypes
        self.integer_columns = integer_columns + data.select_dtypes(include=['int64']).columns.tolist()
        self.mixed_columns = mixed_columns
        for column in integer_columns:
            if column not in self.data_type: 
                warnings.warn(f"Column {column} not found in data, ignoring integer assignment")
                continue
            if data[column].isna().any(): continue # If we have nan values, we cant assign the column to interger, however we still treat it as integer
            data[column] = data[column].astype(int)
        
        self.number_of_decimals = self.get_column_desimal(data)
        

        

    def assign(self, data):
        for column in data.columns:
            if column in self.number_of_decimals and self.number_of_decimals[column] is not None:
                exclude_values = self.mixed_columns.get(column, [])
                data[column] = data[column].apply(
                    lambda x: round(x, self.number_of_decimals[column]) if x not in exclude_values else x
                )
        
        data = data.astype(self.data_type)
        return data

    
    def get_column_desimal(self, data):
        decimals = {}
        for column in data.columns:
            if data[column].dtype != 'int64':
                exclude_values = self.mixed_columns.get(column, [])
                decimals[column] = data[column].apply(
                    lambda x: len(str(x).split('.')[1]) if '.' in str(x) and float(f"0.{str(x).split('.')[1]}") not in exclude_values else 0
                ).max()
            else:
                decimals[column] = 0  # Integer columns have 0 decimals by default

        for column in self.integer_columns:  # Ensure integer columns are explicitly set to 0 decimals
            decimals[column] = 0
        return decimals





from .evaluation import (
    get_summary_metrics,
    get_utility_metrics,
    stat_sim,
    alternative_correlation,
    privacy_metrics,
    compute_nnaa,
    compare_dataframes

)

__all__ = [
    "get_summary_metrics",
    "get_utility_metrics",
    "stat_sim",
    "alternative_correlation",
    "privacy_metrics",
    "compute_nnaa",
    "compare_dataframes"
]
   



from enum import Enum
import numpy as np


## TODO: not used, probabilly could be used when columsn are not specified and we have to check them
class Transform_type(Enum):
    CATEGORICAL = 1
    MIXED = 2
    GAUSSIAN = 3
    CONTINUOUS = 4


class Preprocess_type(Enum):
    CATEGORICAL = 1
    LOG = 2

class Column_assigner:
   


    @classmethod
    def check_column_transform_type(cls, data_col):
        # TODO: One could add some checks for something here at a later point
        # For now we just assume it is continuous and fit by a GMM
        if cls.is_categorical(data_col):
            raise ValueError("Column: ", data_col, "is categorical. Include it in the categorical list")
        return Transform_type.CONTINUOUS

    @classmethod
    def is_categorical(cls, data_col):
        # Implement some fancy check here
        return False
        return False # Implement some fancy check here

                

        

    




from pathlib import Path
import pandas as pd

def load_demo(dataset_name = "adult"):


    dataset_file = get_dataset_path(dataset_name)


    file_path = Path(__file__).resolve()
    folder_path = file_path.parent
    csv_path = folder_path / dataset_file
    
    if not csv_path.exists():
        return csv_path
        raise FileNotFoundError("Demo file not found")
        
    return pd.read_csv(csv_path)


def display_demo():
    datasets = ["adult", "car", "king"]
    return datasets

def get_dataset_path(dataset_name):
    if dataset_name == "adult": return "Adult.csv"
    if dataset_name == "car": return "car.csv"
    if dataset_name == "king": return "king.csv"
    raise ValueError("Invalid dataset name")
    


import numpy as np
import pandas as pd
import torch
import torch.utils.data
import torch.optim as optim
from torch.optim import Adam
from torch.nn import functional as F
from torch.nn import (Dropout, LeakyReLU, Linear, Module, ReLU, Sequential,
Conv2d, ConvTranspose2d, Sigmoid, init, BCELoss, CrossEntropyLoss,SmoothL1Loss,LayerNorm)
from .transformer import ImageTransformer,DataTransformer
from ..privacy_utils.rdp_accountant import compute_rdp, get_privacy_spent
from tqdm import tqdm

from ..privacy_utils.DP_controller import DP_controller




class Classifier(Module):
    def __init__(self,input_dim, dis_dims,st_ed):
        super(Classifier,self).__init__()
        dim = input_dim-(st_ed[1]-st_ed[0])
        seq = []
        self.str_end = st_ed
        for item in list(dis_dims):
            seq += [
                Linear(dim, item),
                LeakyReLU(0.2),
                Dropout(0.5)
            ]
            dim = item
        
        if (st_ed[1]-st_ed[0])==1:
            seq += [Linear(dim, 1)]
        
        elif (st_ed[1]-st_ed[0])==2:
            seq += [Linear(dim, 1),Sigmoid()]
        else:
            seq += [Linear(dim,(st_ed[1]-st_ed[0]))] 
        
        self.seq = Sequential(*seq)

    def forward(self, input):
        
        label=None
        
        if (self.str_end[1]-self.str_end[0])==1:
            label = input[:, self.str_end[0]:self.str_end[1]]
        else:
            label = torch.argmax(input[:, self.str_end[0]:self.str_end[1]], axis=-1)
        
        new_imp = torch.cat((input[:,:self.str_end[0]],input[:,self.str_end[1]:]),1)
        
        if ((self.str_end[1]-self.str_end[0])==2) | ((self.str_end[1]-self.str_end[0])==1):
            return self.seq(new_imp).view(-1), label
        else:
            return self.seq(new_imp), label

def apply_activate(data, output_info):
    data_t = []
    st = 0
    for item in output_info:
        if item[1] == 'tanh':
            ed = st + item[0]
            data_t.append(torch.tanh(data[:, st:ed]))
            st = ed
        elif item[1] == 'softmax':
            ed = st + item[0]
            data_t.append(F.gumbel_softmax(data[:, st:ed], tau=0.2))
            st = ed
    return torch.cat(data_t, dim=1)

def get_st_ed(target_col_index,output_info):
    st = 0
    c= 0
    tc= 0

    for item in output_info:
        if c==target_col_index:
            break
        if item[1]=='tanh':
            st += item[0]
            if item[2] == 'yes_g':
                c+=1
        elif item[1] == 'softmax':
            st += item[0]
            c+=1
        tc+=1    
    
    ed= st+output_info[tc][0] 

    return (st,ed)

def random_choice_prob_index_sampling(probs,col_idx):
    option_list = []
    for i in col_idx:
        pp = probs[i]
        option_list.append(np.random.choice(np.arange(len(probs[i])), p=pp))
    
    return np.array(option_list).reshape(col_idx.shape)

def random_choice_prob_index(a, axis=1):
    r = np.expand_dims(np.random.rand(a.shape[1 - axis]), axis=axis)
    return (a.cumsum(axis=axis) > r).argmax(axis=axis)

def maximum_interval(output_info):
    max_interval = 0
    for item in output_info:
        max_interval = max(max_interval, item[0])
    return max_interval

class Cond(object):
    def __init__(self, data, output_info):
       
        self.model = []
        st = 0
        counter = 0
        for item in output_info:
           
            if item[1] == 'tanh':
                st += item[0]
                continue
            elif item[1] == 'softmax':
                ed = st + item[0]
                counter += 1
                self.model.append(np.argmax(data[:, st:ed], axis=-1))
                st = ed
            
        self.interval = []
        self.n_col = 0  
        self.n_opt = 0  
        st = 0
        self.p = np.zeros((counter, maximum_interval(output_info)))  
        self.p_sampling = []
        for item in output_info:
            if item[1] == 'tanh':
                st += item[0]
                continue
            elif item[1] == 'softmax': 
                ed = st + item[0]
                data_cols = np.array(data[:, st:ed],np.float64)
                tmp = np.sum(data_cols, axis=0)  
                tmp = np.log(tmp + 1)  
                tmp = tmp / np.sum(tmp) 
                tmp_sampling = np.sum(data_cols, axis=0)     
                tmp_sampling = tmp_sampling / np.sum(tmp_sampling)
                self.p_sampling.append(tmp_sampling)
                self.p[self.n_col, :item[0]] = tmp 
                self.interval.append((self.n_opt, item[0]))
                self.n_opt += item[0]
                self.n_col += 1
                st = ed
                
        self.interval = np.asarray(self.interval)
        
    def sample_train(self, batch):
        if self.n_col == 0:
            return None
        batch = batch

        idx = np.random.choice(np.arange(self.n_col), batch)

        vec = np.zeros((batch, self.n_opt), dtype='float32')
        mask = np.zeros((batch, self.n_col), dtype='float32')
        mask[np.arange(batch), idx] = 1  
        opt1prime = random_choice_prob_index(self.p[idx]) 
        for i in np.arange(batch):
            vec[i, self.interval[idx[i], 0] + opt1prime[i]] = 1
            
        return vec, mask, idx, opt1prime

    def sample(self, batch):
        if self.n_col == 0:
            return None
        batch = batch
      
        idx = np.random.choice(np.arange(self.n_col), batch)

        vec = np.zeros((batch, self.n_opt), dtype='float32')
        opt1prime = random_choice_prob_index_sampling(self.p_sampling,idx)
        
        for i in np.arange(batch):
            vec[i, self.interval[idx[i], 0] + opt1prime[i]] = 1
            
        return vec

def cond_loss(data, output_info, c, m):
    loss = []
    st = 0
    st_c = 0
    for item in output_info:
        if item[1] == 'tanh':
            st += item[0]
            continue

        elif item[1] == 'softmax':
            ed = st + item[0]
            ed_c = st_c + item[0]
            tmp = F.cross_entropy(
            data[:, st:ed],
            torch.argmax(c[:, st_c:ed_c], dim=1),
            reduction='none')
            loss.append(tmp)
            st = ed
            st_c = ed_c

    loss = torch.stack(loss, dim=1)
    return (loss * m).sum() / data.size()[0]

class Sampler(object):
    def __init__(self, data, output_info):
        super(Sampler, self).__init__()
        self.data = data
        self.model = []
        self.n = len(data)
        st = 0
        for item in output_info:
            if item[1] == 'tanh':
                st += item[0]
                continue
            elif item[1] == 'softmax':
                ed = st + item[0]
                tmp = []
                for j in range(item[0]):
                    tmp.append(np.nonzero(data[:, st + j])[0])
                self.model.append(tmp)
                st = ed
                
    def sample(self, n, col, opt):
        if col is None:
            idx = np.random.choice(np.arange(self.n), n)
            return self.data[idx]
        idx = []
        for c, o in zip(col, opt):
            idx.append(np.random.choice(self.model[c][o]))
        return self.data[idx]

class Discriminator(Module):
    def __init__(self, side, layers):
        super(Discriminator, self).__init__()
        self.side = side
        info = len(layers)-2
        self.seq = Sequential(*layers)
        self.seq_info = Sequential(*layers[:info])

    def forward(self, input):
        return (self.seq(input)), self.seq_info(input)

class Generator(Module):
    def __init__(self, side, layers):
        super(Generator, self).__init__()
        self.side = side
        self.seq = Sequential(*layers)

    def forward(self, input_):
        return self.seq(input_)

def determine_layers_disc(side, num_channels):
    assert side >= 4 and side <= 64

    layer_dims = [(1, side), (num_channels, side // 2)]

    while layer_dims[-1][1] > 3 and len(layer_dims) < 4:
        layer_dims.append((layer_dims[-1][0] * 2, layer_dims[-1][1] // 2))

    layerNorms = []
    num_c = num_channels
    num_s = side / 2
    for l in range(len(layer_dims) - 1):
        layerNorms.append([int(num_c), int(num_s), int(num_s)])
        num_c = num_c * 2
        num_s = num_s / 2

    layers_D = []

    for prev, curr, ln in zip(layer_dims, layer_dims[1:], layerNorms):
        layers_D += [
            Conv2d(prev[0], curr[0], 4, 2, 1, bias=False),
            LayerNorm(ln),
            LeakyReLU(0.2, inplace=True),
        ]

    layers_D += [Conv2d(layer_dims[-1][0], 1, layer_dims[-1][1], 1, 0), ReLU(True)] 

    return layers_D

def determine_layers_gen(side, random_dim, num_channels):
    assert side >= 4 and side <= 64

    layer_dims = [(1, side), (num_channels, side // 2)]

    while layer_dims[-1][1] > 3 and len(layer_dims) < 4:
        layer_dims.append((layer_dims[-1][0] * 2, layer_dims[-1][1] // 2))

    layerNorms = []

    num_c = num_channels * (2 ** (len(layer_dims) - 2))
    num_s = int(side / (2 ** (len(layer_dims) - 1)))
    for l in range(len(layer_dims) - 1):
        layerNorms.append([int(num_c), int(num_s), int(num_s)])
        num_c = num_c / 2
        num_s = num_s * 2

    layers_G = [ConvTranspose2d(random_dim, layer_dims[-1][0], layer_dims[-1][1], 1, 0, output_padding=0, bias=False)]

    for prev, curr, ln in zip(reversed(layer_dims), reversed(layer_dims[:-1]), layerNorms):
        layers_G += [LayerNorm(ln), ReLU(True), ConvTranspose2d(prev[0], curr[0], 4, 2, 1, output_padding=0, bias=True)]
    return layers_G

def slerp(val, low, high):
    low_norm = low/torch.norm(low, dim=1, keepdim=True)
    high_norm = high/torch.norm(high, dim=1, keepdim=True)
    omega = torch.acos((low_norm*high_norm).sum(1)).view(val.size(0), 1)
    so = torch.sin(omega)
    res = (torch.sin((1.0-val)*omega)/so)*low + (torch.sin(val*omega)/so) * high
    
    return res

def calc_gradient_penalty_slerp(netD, real_data, fake_data, transformer, device='cpu', lambda_=10):
    batchsize = real_data.shape[0]
    alpha = torch.rand(batchsize, 1,  device=device)
    interpolates = slerp(alpha, real_data, fake_data)
    interpolates = interpolates.to(device)
    interpolates = transformer.transform(interpolates)
    interpolates = torch.autograd.Variable(interpolates, requires_grad=True)
    disc_interpolates,_ = netD(interpolates) 

    gradients = torch.autograd.grad(outputs=disc_interpolates, inputs=interpolates,
                                  grad_outputs=torch.ones(disc_interpolates.size()).to(device),
                                  create_graph=True, retain_graph=True, only_inputs=True)[0] 
    
    gradients_norm = gradients.norm(2, dim=1)
    gradient_penalty = ((gradients_norm - 1) ** 2).mean() * lambda_
    
    return gradient_penalty

def weights_init(m):
    classname = m.__class__.__name__
    
    if classname.find('Conv') != -1:
        init.normal_(m.weight.data, 0.0, 0.02)

    elif classname.find('BatchNorm') != -1:
        init.normal_(m.weight.data, 1.0, 0.02)
        init.constant_(m.bias.data, 0)

class CTABGANSynthesizer:
    def __init__(self,
                 class_dim=(256, 256, 256, 256),
                 random_dim=100,
                 num_channels=64,
                 l2scale=1e-5,
                 batch_size=500):
                 

        self.random_dim = random_dim
        self.class_dim = class_dim
        self.num_channels = num_channels
        self.dside = None
        self.gside = None
        self.l2scale = l2scale
        self.batch_size = batch_size

        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    def fit(self, train_data=pd.DataFrame, dp_constraints={}, categorical=[], mixed={}, general=[], non_categorical=[], type={},epochs=100,verbose=True):

        self.epochs = epochs
        problem_type = None
        target_index=None
        if type:
            problem_type = list(type.keys())[0]
            if problem_type:
                target_index = train_data.columns.get_loc(type[problem_type])

        self.transformer = DataTransformer(train_data, categorical_list=categorical, mixed_dict=mixed, general_list=general)
        self.transformer.fit() 
        #output_info = self.transformer.get_output_info_flat()
        output_info = self.transformer.output_info
        train_data = self.transformer.transform(train_data.values)
        data_sampler = Sampler(train_data, output_info)
        data_dim = self.transformer.output_dim
        self.cond_generator = Cond(train_data, output_info)
        		
        sides = [4, 8, 16, 24, 32, 64]
        col_size_d = data_dim + self.cond_generator.n_opt
        for i in sides:
            if i * i >= col_size_d:
                self.dside = i
                break
        
        sides = [4, 8, 16, 24, 32, 64]
        col_size_g = data_dim
        for i in sides:
            if i * i >= col_size_g:
                self.gside = i
                break
		

        layers_G = determine_layers_gen(self.gside, self.random_dim+self.cond_generator.n_opt, self.num_channels)
        layers_D = determine_layers_disc(self.dside, self.num_channels)
        
        self.generator = Generator(self.gside, layers_G).to(self.device)
        discriminator = Discriminator(self.dside, layers_D).to(self.device)
        optimizer_params = dict(lr=2e-4, betas=(0.5, 0.9), eps=1e-3, weight_decay=self.l2scale)
        optimizerG = Adam(self.generator.parameters(), **optimizer_params)
        optimizerD = Adam(discriminator.parameters(), **optimizer_params)

        st_ed = None
        classifier=None
        optimizerC= None
        if target_index != None:
            st_ed= get_st_ed(target_index,output_info)
            classifier = Classifier(data_dim,self.class_dim,st_ed).to(self.device)
            optimizerC = optim.Adam(classifier.parameters(),**optimizer_params)
        
        
        self.generator.apply(weights_init)
        discriminator.apply(weights_init)

        self.Gtransformer = ImageTransformer(self.gside)       
        self.Dtransformer = ImageTransformer(self.dside)
        
        epsilon = 0
        epoch = 0
        steps = 0
        ci = 5
        
        steps_per_epoch = max(1, len(train_data) // self.batch_size)
        total_steps = steps_per_epoch * epochs * ci

        self.private = any(value is not None for value in dp_constraints.values())
        
        if self.private:
             
            self.micro_batch_size = self.batch_size
            
            self.dp_controller = DP_controller(
                        data_size=train_data.shape[0], 
                        batch_size=self.batch_size, 
                        total_steps=total_steps, 
                        epsilon_budget=dp_constraints.get('epsilon_budget',None), 
                        delta=dp_constraints.get('delta',None), 
                        clip_coeff=dp_constraints.get('clip_coeff',1), 
                        sigma=dp_constraints.get('sigma',None), 
                        device=self.device, 
                        verbose=verbose)

            self.clip_coeff = self.dp_controller.get_clip_coeff()
            self.sigma = self.dp_controller.get_sigma()




        for i in tqdm(range(self.epochs)):
            for id_ in range(steps_per_epoch):
				
                
                for _ in range(ci):
                    noisez = torch.randn(self.batch_size, self.random_dim, device=self.device)
                    condvec = self.cond_generator.sample_train(self.batch_size)

                    c, m, col, opt = condvec
                    c = torch.from_numpy(c).to(self.device)
                    m = torch.from_numpy(m).to(self.device)
                    noisez = torch.cat([noisez, c], dim=1)
                    noisez =  noisez.view(self.batch_size,self.random_dim+self.cond_generator.n_opt,1,1)
                    
                    perm = np.arange(self.batch_size)
                    np.random.shuffle(perm)
                    real = data_sampler.sample(self.batch_size, col[perm], opt[perm])
                    c_perm = c[perm]
                    
                    real = torch.from_numpy(real.astype('float32')).to(self.device)
                    
                    fake = self.generator(noisez)
                    faket = self.Gtransformer.inverse_transform(fake)
                    fakeact = apply_activate(faket, output_info)
                    
                    fake_cat = torch.cat([fakeact, c], dim=1)
                    real_cat = torch.cat([real, c_perm], dim=1)
                    
                    real_cat_d = self.Dtransformer.transform(real_cat)
                    fake_cat_d = self.Dtransformer.transform(fake_cat)
                    
                    optimizerD.zero_grad()
                    
                    d_real,_ = discriminator(real_cat_d)
                    

                    if self.private: #  IF private we have to apply the gradient clipping and the noise addition

                        clipped_grads = {
                            name: torch.zeros_like(param) for name, param in discriminator.named_parameters()}

                        for k in range(int(d_real.size(0) / self.micro_batch_size)):
                            err_micro = -1*d_real[k * self.micro_batch_size: (k + 1) * self.micro_batch_size].mean(0).view(1)
                            err_micro.backward(retain_graph=True)
                            torch.nn.utils.clip_grad_norm_(discriminator.parameters(), self.clip_coeff)
                            for name, param in discriminator.named_parameters():
                                clipped_grads[name] += param.grad
                            discriminator.zero_grad()

                        for name, param in discriminator.named_parameters():
                            param.grad = (clipped_grads[name] + torch.FloatTensor(
                                clipped_grads[name].size()).normal_(0, self.sigma * self.clip_coeff).to(self.device)) / (
                                                     d_real.size(0) / self.micro_batch_size)
                        steps += 1
                    else: 
                        d_real = -torch.mean(d_real)
                        d_real.backward() 
                    

                    d_fake,_ = discriminator(fake_cat_d)
                    
                    d_fake = torch.mean(d_fake)

                    d_fake.backward() 
                    
                    pen = calc_gradient_penalty_slerp(discriminator, real_cat, fake_cat,  self.Dtransformer , self.device)

                    pen.backward()
                
                    optimizerD.step()
                    
                noisez = torch.randn(self.batch_size, self.random_dim, device=self.device)
                
                condvec = self.cond_generator.sample_train(self.batch_size)

                c, m, col, opt = condvec
                c = torch.from_numpy(c).to(self.device)
                m = torch.from_numpy(m).to(self.device)
                noisez = torch.cat([noisez, c], dim=1)
                noisez =  noisez.view(self.batch_size,self.random_dim+self.cond_generator.n_opt,1,1)

                optimizerG.zero_grad()

                fake = self.generator(noisez)
                faket = self.Gtransformer.inverse_transform(fake)
                fakeact = apply_activate(faket, output_info)

                fake_cat = torch.cat([fakeact, c], dim=1) 
                fake_cat = self.Dtransformer.transform(fake_cat)
                    
                y_fake,info_fake = discriminator(fake_cat)
                
                cross_entropy = cond_loss(faket, output_info, c, m)

                _,info_real = discriminator(real_cat_d)
                

                g = -torch.mean(y_fake) + cross_entropy
                g.backward(retain_graph=True)
                loss_mean = torch.norm(torch.mean(info_fake.view(self.batch_size,-1), dim=0) - torch.mean(info_real.view(self.batch_size,-1), dim=0), 1)
                loss_std = torch.norm(torch.std(info_fake.view(self.batch_size,-1), dim=0) - torch.std(info_real.view(self.batch_size,-1), dim=0), 1)
                loss_info = loss_mean + loss_std 
                loss_info.backward()
                optimizerG.step()


                if problem_type:
                           
                    fake = self.generator(noisez)
                    
                    faket = self.Gtransformer.inverse_transform(fake)
                    
                    fakeact = apply_activate(faket, output_info)
                    
                    real_pre, real_label = classifier(real)
                    fake_pre, fake_label = classifier(fakeact)
                     
                    c_loss = CrossEntropyLoss() 
                    
                    if (st_ed[1] - st_ed[0])==1:
                        c_loss= SmoothL1Loss()
                        real_label = real_label.type_as(real_pre)
                        fake_label = fake_label.type_as(fake_pre)
                        real_label = torch.reshape(real_label,real_pre.size())
                        fake_label = torch.reshape(fake_label,fake_pre.size())
                        
                    
                    elif (st_ed[1] - st_ed[0])==2:
                        c_loss = BCELoss()
                        real_label = real_label.type_as(real_pre)
                        fake_label = fake_label.type_as(fake_pre)

                    loss_cc = c_loss(real_pre, real_label)
                    loss_cg = c_loss(fake_pre, fake_label)

                    optimizerG.zero_grad()
                    loss_cg.backward()
                    optimizerG.step()

                    optimizerC.zero_grad()
                    loss_cc.backward()
                    optimizerC.step()
                                
            epoch += 1
            if self.private:
                
                epsilon = self.dp_controller.get_spent_privacy(steps)
                if verbose: print("Epoch :", epoch, "Epsilon spent : ", epsilon)

            
   
    def sample(self, n):
        
        self.generator.eval()

        output_info = self.transformer.output_info
        steps = n // self.batch_size + 1
        
        data = []
        
        for i in range(steps):
            noisez = torch.randn(self.batch_size, self.random_dim, device=self.device)
            condvec = self.cond_generator.sample(self.batch_size)
            c = condvec
            c = torch.from_numpy(c).to(self.device)
            noisez = torch.cat([noisez, c], dim=1)
            noisez =  noisez.view(self.batch_size,self.random_dim+self.cond_generator.n_opt,1,1)
                
            fake = self.generator(noisez)
            faket = self.Gtransformer.inverse_transform(fake)
            fakeact = apply_activate(faket,output_info)
            data.append(fakeact.detach().cpu().numpy())

        data = np.concatenate(data, axis=0)
        result,resample = self.transformer.inverse_transform(data)
        
        while len(result) < n:
            data_resample = []    
            steps_left = resample// self.batch_size + 1
            
            for i in range(steps_left):
                noisez = torch.randn(self.batch_size, self.random_dim, device=self.device)
                condvec = self.cond_generator.sample(self.batch_size)
                c = condvec
                c = torch.from_numpy(c).to(self.device)
                noisez = torch.cat([noisez, c], dim=1)
                noisez =  noisez.view(self.batch_size,self.random_dim+self.cond_generator.n_opt,1,1)
                    
                fake = self.generator(noisez)
                faket = self.Gtransformer.inverse_transform(fake)
                fakeact = apply_activate(faket, output_info)
                data_resample.append(fakeact.detach().cpu().numpy())

            data_resample = np.concatenate(data_resample, axis=0)

            res,resample = self.transformer.inverse_transform(data_resample)
            result  = np.concatenate([result,res],axis=0)
        
        return result[0:n]

